package spark.listnet

import spark.SparkContext
import SparkContext._

import java.util.Random
import spark.examples.Vector._

import scala.tools.nsc.interpreter.ILoop._

object GradientDescentSerial {
	// Numer of dimensions in the feature vector
	val dim = 136
	val rand = new Random(42)
  
  	// Data Structure that contains query id, 
  	// relevance score and feature vector 
  	case class QueryDS(qid: String, relScores: Array[Int], 
  		docFeatures: Array[spark.examples.Vector])
    
    // Preprocessed data is of the form qid:<qid>#rel1 
    // 1:<1> 2:<2>....136:<136>#rel2 1:<1>....136:<136>
    // parseQuery populates QueryDS by parsing the preprocessed data
  	def parseQuery(line: String): QueryDS = {
    	val tokens = line.split('#')
    	val qid = tokens(0)
    	val relScores = tokens.slice(1, tokens.length).map(_.split(" ")(0).toInt)
    	val docFeatures = tokens.slice(1, tokens.length)
    		.map(document => spark.examples.Vector(document.split(" ").slice(1, 137).map(_.split(":")(1).toDouble)))
    	return QueryDS(qid, relScores, docFeatures)
  	}
  
	def main( args: Array[String]) {
	
		if (args.length != 4) {
      		System.err.println("Usage: SparkTest <host> <filename> <num iterations> <step size>")
      		System.exit(1)
    	}
		var startTime = System.currentTimeMillis();	
		
		//val sparkContext = new SparkContext(args(0), "SparkTest")
		//val myFile = sparkContext.textFile(args(1))
		
		val myFile = scala.io.Source.fromFile(args(1)).getLines().toList
		
		//breakIf(true,"myFile" -> myFile)
		
		// queries is a list of QueryDSs
		val queries = myFile.map(line => parseQuery(line))
		// println("queries.length.toDouble =" + queries.length.toDouble)
		
		val ITERATIONS = args(2).toInt
		
		val stepSize = args(3).toDouble

		// Initialize weight vector w to a random value
		var w = spark.examples.Vector.zeros(dim)
		
		var gradient = spark.examples.Vector.zeros(dim)
		
		var loss = 0.0
		
		println("Initial w: " + w)
		
		for (i <- 1 to ITERATIONS) {
			println("On iteration " + i)
			gradient = spark.examples.Vector.zeros(dim)
			loss = 0.0
			 
			for (q <- queries) {
				
				// Exponentiate all relevance scores
				val expRelScores = q.relScores.map(y => math.exp(y.toDouble))
				//println(" expRelScores(0)" + expRelScores(0))
				// Exponentiate our scores calculated 
				// by taking a dot product of feature vector with weight vector
				val ourScores = q.docFeatures.map(x => w dot x)
				//println("ourScores" + ourScores.mkString(" "))
				
				val expOurScores = ourScores.map(z => math.exp(z))
				//println("expOurScores" + expOurScores.mkString(" "))
				
				// Calculate sum of all the relevance 
				// scores and scores calculated with our weight vector
				val sumExpRelScores = expRelScores.reduce(_ + _)
				val sumExpOurScores = expOurScores.reduce(_ + _)
				
				val P_y = expRelScores.map(y => y/sumExpRelScores)
				val P_z = expOurScores.map(z => z/sumExpOurScores)
				
				var lossForAQuery = 0.0
				
				//println( " P_y(0) = " + P_y(0))
				var gradientForAQuery = spark.examples.Vector.zeros(dim)
				for (j <- 0 to q.relScores.length-1) {
					/*
					println("q.docFeatures(j)" + q.docFeatures(j))
					println( " P_y(j) = " + P_y(j))
					println( " P_z(j) = " + P_z(j))
					println( " P_z(j) - P_y(j) = " + (P_z(j) - P_y(j)))
					println( " q.docFeatures(j) * (P_z(j) - P_y(j)) = " + (q.docFeatures(j) * (P_z(j) - P_y(j))))
					*/
					
					lossForAQuery += -P_y(j) * math.log(P_z(j))
					gradientForAQuery += (q.docFeatures(j) * (P_z(j) - P_y(j)))
					//println("gradientForAQuery(0) = " + gradientForAQuery(0))
				}
				//println("gradientForAQuery = " + gradientForAQuery)
				
				gradient += gradientForAQuery
				loss += lossForAQuery
				//w -= gradientForAQuery
				//println("inner w = " + w)
			}
			println("Current loss = " + loss)
			
			w -= gradient * stepSize
			//println("outer w = " + w)			
		}
		println("Final w: " + w)
		println("\n Job Finished in " + (System.currentTimeMillis() - startTime) / 1000.0 + " seconds")
	}
}
