package spark.listnet

import spark._
import SparkContext._
import spark.examples.Vector._
import scala.io.Source
import java.util.Random

object NDCGCalculator {
	// Numer of dimensions in the feature vector
	val dim = 136
 	val rand = new Random(42)
 	
  	// Data Structure that contains query id, 
  	// relevance score and feature vector 
  	case class QueryDS(qid: String, relScores: Array[Int], 
  		docFeatures: Array[spark.examples.Vector])
    
    // Preprocessed data is of the form qid:<qid>#rel1 
    // 1:<1> 2:<2>....136:<136>#rel2 1:<1>....136:<136>
    // parseQuery populates QueryDS by parsing the preprocessed data
  	def parseQuery(line: String): QueryDS = {
    	val tokens = line.split('#')
    	val qid = tokens(0)
    	val relScores = tokens.slice(1, tokens.length).map(_.split(" ")(0).toInt)
    	val docFeatures = tokens.slice(1, tokens.length)
    		.map(document => spark.examples.Vector(document.split(" ").slice(1, 137).map(_.split(":")(1).toDouble)))
    	return QueryDS(qid, relScores, docFeatures)
  	}

	
	def main(args: Array[String]) {
		val dim = 136
		if (args.length != 3) {
      		System.err.println("Usage: SparkTest <host> <data filename> <weight vector file>")
      		System.exit(1)
    	}
		var startTime = System.currentTimeMillis();		
		
		val sparkContext = new SparkContext(args(0), "SparkTest")
		val myFile = sparkContext.textFile(args(1))
		
		// queries is a list of QueryDSs
		val queries = myFile.map(line => parseQuery(line)).cache()	
		
		// Initialize weight vector w with zeros
		val weightVectorString = Source.fromFile(args(2)).mkString
		//val w = new spark.examples.Vector(weightVectorString.split(",").map(x=>x.toDouble))

		var w = spark.examples.Vector(136, _ => 2 * rand.nextDouble - 1)
		println(w)
		
		val NDCG = sparkContext.accumulator(0.0)
		val queryCount = sparkContext.accumulator(0)
		
		for (q <- queries) {
			val ourScores = q.docFeatures.map(x => w dot x)
			val ourRankedScores = ourScores.zip(0 to ourScores.length-1)
			val ourSortedRankedScores = ourRankedScores.sortBy(x => -x._1)
			var DCG4AQuery = 0.0
			for (j <- 0 to math.min(10, ourSortedRankedScores.length-1)) {
			//for (j <- 0 to ourSortedRankedScores.length-1) {	
				DCG4AQuery += (math.pow(2, q.relScores(ourSortedRankedScores(j)._2)) - 1)/math.max(math.log(j+1)/math.log(2),1 )
			}
			
			val sortedRelScores = q.relScores.sortBy(x => -x)
			var MaxDCG4AQuery = 0.0
			for (j <- 0 to math.min(10, sortedRelScores.length-1) ){
			//for (j <- 0 to sortedRelScores.length-1) {		
				MaxDCG4AQuery += (math.pow(2, sortedRelScores(j)) - 1)/math.max(math.log(j+1)/math.log(2),1 )
			}
			val NDCG4AQuery = if ( MaxDCG4AQuery == 0 ) 0 else DCG4AQuery / MaxDCG4AQuery
			
			NDCG += NDCG4AQuery
			queryCount += 1
			//println("NDCG for query = " + NDCG4AQuery + " MaxDCG4AQuery = " + MaxDCG4AQuery + " DCG4AQuery = " + DCG4AQuery)
		}
		println("average NDCG = " + NDCG.value/queryCount.value)
		println("\n Job Finished in " + (System.currentTimeMillis() - startTime) / 1000.0 + " seconds")
	}
}
